import json
import os
import sys
import argparse
import pathlib
import shutil

from jupyter_client.kernelspec import KernelSpecManager
from IPython.utils.tempdir import TemporaryDirectory

# from .resources import _ICON_PATH

kernel_json = {
    "argv": [sys.executable, "-m", "q8s", "-f", "{connection_file}"],
    "display_name": "Q8s kernel",
    "language": "python",
}


def install_my_kernel_spec(user=True, prefix=None):
    with TemporaryDirectory() as td:
        os.chmod(td, 0o755)  # Starts off as 700, not user readable
        with open(os.path.join(td, "kernel.json"), "w") as f:
            json.dump(kernel_json, f, sort_keys=True)
        # shutil.copyfile(_ICON_PATH, pathlib.Path(td) / _ICON_PATH.name)
        print("Installing q8s kernel spec")
        KernelSpecManager().install_kernel_spec(td, "q8s", user=user, prefix=prefix)


def _is_root():
    try:
        return os.geteuid() == 0
    except AttributeError:
        return False  # assume not an admin on non-Unix platforms


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Install KernelSpec for Qubernetes Kernel"
    )
    prefix_locations = parser.add_mutually_exclusive_group()

    prefix_locations.add_argument(
        "--user",
        help="Install KernelSpec in user's home directory",
        action="store_true",
    )
    prefix_locations.add_argument(
        "--sys-prefix",
        help="Install KernelSpec in sys.prefix. Useful in conda / virtualenv",
        action="store_true",
        dest="sys_prefix",
    )
    prefix_locations.add_argument(
        "--prefix", help="Install KernelSpec in this prefix", default=None
    )

    args = parser.parse_args(argv)

    user = False
    prefix = None
    if args.sys_prefix:
        prefix = sys.prefix
    elif args.prefix:
        prefix = args.prefix
    elif args.user or not _is_root():
        user = True

    install_my_kernel_spec(user=user, prefix=prefix)


if __name__ == "__main__":
    main()



import unittest
from unittest.mock import patch, MagicMock
from kubernetes import client
from q8s.plugins.job_template_spec import CPUJobTemplatePlugin, CUDAJobTemplatePlugin
from q8s.enums import Target


class TestCPUandGPUJobTemplatePlugin(unittest.TestCase):

    @patch("q8s.plugins.job_template_spec.client.V1Container")
    @patch("q8s.plugins.job_template_spec.client.V1PodTemplateSpec")
    @patch("q8s.plugins.job_template_spec.client.V1PodSpec")
    @patch("q8s.plugins.job_template_spec.client.V1ObjectMeta")
    @patch("q8s.plugins.job_template_spec.client.V1VolumeMount")
    @patch("q8s.plugins.job_template_spec.client.V1Volume")
    @patch("q8s.plugins.job_template_spec.client.V1ConfigMapVolumeSource")
    @patch("q8s.plugins.job_template_spec.client.V1LocalObjectReference")
    @patch("q8s.plugins.job_template_spec.client.V1ResourceRequirements")
    def test_makejob_cpu(
        self,
        mock_v1_resource_requirements,
        mock_v1_local_object_reference,
        mock_v1_config_map_volume_source,
        mock_v1_volume,
        mock_v1_volume_mount,
        mock_v1_object_meta,
        mock_v1_pod_spec,
        mock_v1_pod_template_spec,
        mock_v1_container,
    ):
        plugin = CPUJobTemplatePlugin()
        name = "test-job"
        registry_pat = None
        registry_credentials_secret_name = "test-secret"
        container_image = "test-image"
        env = {"TEST_ENV": "value"}
        target = Target.cpu

        result = plugin.makejob(
            name,
            registry_pat,
            registry_credentials_secret_name,
            container_image,
            env,
            target,
        )

        self.assertIsNotNone(result)
        mock_v1_container.assert_called_once()
        mock_v1_pod_template_spec.assert_called_once()
        mock_v1_pod_spec.assert_called_once()
        mock_v1_object_meta.assert_called_once()
        mock_v1_volume_mount.assert_called_once()
        mock_v1_volume.assert_called_once()
        mock_v1_config_map_volume_source.assert_called_once()

    @patch("q8s.plugins.job_template_spec.client.V1Container")
    @patch("q8s.plugins.job_template_spec.client.V1PodTemplateSpec")
    @patch("q8s.plugins.job_template_spec.client.V1PodSpec")
    @patch("q8s.plugins.job_template_spec.client.V1ObjectMeta")
    @patch("q8s.plugins.job_template_spec.client.V1VolumeMount")
    @patch("q8s.plugins.job_template_spec.client.V1Volume")
    @patch("q8s.plugins.job_template_spec.client.V1ConfigMapVolumeSource")
    @patch("q8s.plugins.job_template_spec.client.V1LocalObjectReference")
    @patch("q8s.plugins.job_template_spec.client.V1ResourceRequirements")
    def test_makejob_gpu(
        self,
        mock_v1_resource_requirements,
        mock_v1_local_object_reference,
        mock_v1_config_map_volume_source,
        mock_v1_volume,
        mock_v1_volume_mount,
        mock_v1_object_meta,
        mock_v1_pod_spec,
        mock_v1_pod_template_spec,
        mock_v1_container,
    ):
        plugin = CUDAJobTemplatePlugin()
        name = "test-job"
        registry_pat = "test-pat"
        registry_credentials_secret_name = "test-secret"
        container_image = "test-image"
        env = {"TEST_ENV": "value"}
        target = Target.gpu

        result = plugin.makejob(
            name,
            registry_pat,
            registry_credentials_secret_name,
            container_image,
            env,
            target,
        )

        self.assertIsNotNone(result)
        mock_v1_container.assert_called_once()
        mock_v1_pod_template_spec.assert_called_once()
        mock_v1_pod_spec.assert_called_once()
        mock_v1_object_meta.assert_called_once()
        mock_v1_volume_mount.assert_called_once()
        mock_v1_volume.assert_called_once()
        mock_v1_config_map_volume_source.assert_called_once()
        mock_v1_local_object_reference.assert_called_once()
        mock_v1_resource_requirements.assert_called_once()

    def test_makejob_invalid_target(self):
        plugin = CPUJobTemplatePlugin()
        name = "test-job"
        registry_pat = None
        registry_credentials_secret_name = "test-secret"
        container_image = "test-image"
        env = {"TEST_ENV": "value"}
        target = "invalid-target"

        result = plugin.makejob(
            name,
            registry_pat,
            registry_credentials_secret_name,
            container_image,
            env,
            target,
        )

        self.assertIsNone(result)


if __name__ == "__main__":
    unittest.main()



import os
from time import time
from ipykernel.kernelbase import Kernel
import logging

from q8s.kernel import Q8sKernel

from .k8s import execute

CODE = ["test_function", "import json; json.dumps(test_function)"]


# handles the testbook get protocol
def do_execute(
    self: Kernel,
    code,
    silent,
    store_history=True,
    user_expressions=None,
    allow_stdin=False,
):

    logging.debug(f"Executing code:\n{code}")

    if code == "test_function":
        stream_content = {
            "name": "stdout",
            "text": "<function __main__.test_function()>",
        }
        self.send_response(self.iopub_socket, "stream", stream_content)
    elif code.startswith("\ntest_function("):
        logging.debug("execute test_function")
        start = time()
        output, stream_name = execute(
            self.code + "\nprint(" + code.strip() + ")",
            None,
            os.environ["DOCKER_IMAGE"],
        )

        logging.debug(output)
        logging.debug(stream_name)
        self.output = output

        stream_content = {
            "name": stream_name,
            # "text": output + f"\nExecution time: {time() - start:.2f} seconds",
            "text": output,
        }
        # self.send_response(self.iopub_socket, "stream", stream_content)
        self.send_response(
            self.iopub_socket,
            "execute_result",
            {
                "data": {
                    "application/json": {
                        "value": output,
                    }
                },
                "metadata": {},
                "execution_count": self.execution_count,
            },
        )
    elif code.startswith("\nimport json"):
        self.send_response(
            self.iopub_socket,
            "execute_result",
            {
                "data": {
                    "application/json": {
                        "value": float(self.output),
                    }
                },
                "metadata": {},
                "execution_count": self.execution_count,
            },
        )
    elif code.startswith("import json; json.dumps(test_function)"):
        stream_content = {
            "name": "stderr",
            "text": "TypeError: Object of type function is not JSON serializable",
        }
        self.send_response(self.iopub_socket, "stream", stream_content)
    else:
        self.code = code

        stream_content = {"name": "stdout", "text": "Code saved"}
        self.send_response(self.iopub_socket, "stream", stream_content)

    logging.info("Execution complete")
    return {
        "status": "ok",
        # The base class increments the execution count
        "execution_count": self.execution_count,
        "payload": [],
        "user_expressions": {},
    }


if os.environ.get("Q8S_BENCHMARK") == "1":
    Q8sKernel.do_execute = do_execute



import os
from pathlib import Path
from time import sleep
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
import sys
from typing_extensions import Annotated
from q8s.execution import K8sContext
from q8s.enums import Target
from q8s.install import install_my_kernel_spec
from q8s.project import Project

app = typer.Typer()


@app.command()
def build(
    init: Annotated[bool, typer.Option(help="Initialize project")] = False,
    target: Annotated[
        Target, typer.Option(help="Execution target", case_sensitive=False)
    ] = None,
    dry_run: Annotated[
        bool, typer.Option(help="Dry run does not push images to the registry")
    ] = False,
    silent: Annotated[bool, typer.Option(help="Silent mode")] = True,
):

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        expand=True,
    ) as progress:
        task = progress.add_task(description="[cyan]Loading project...", total=1)

        project = Project()

        progress.advance(task)

        if init:
            task = progress.add_task(description="[cyan]Initializing cache...", total=1)

            project.init_cache()
            progress.advance(task)

        if target:
            project.build_container(
                target=target.value,
                progress=progress,
                push=(not dry_run),
                silent=silent,
            )

        else:
            for build in project.configuration.targets.keys():
                project.build_container(
                    build, progress=progress, push=(not dry_run), silent=silent
                )

    print(f"Project {project.name} ready")
    project.update_images_cache()


@app.command()
def execute(
    file: Annotated[Path, typer.Argument(help="Python file to be executed")],
    target: Annotated[
        Target, typer.Option(help="Execution target", case_sensitive=False)
    ] = Target.gpu,
    kubeconfig: Annotated[
        Path, typer.Option(help="Kubernetes configuration", envvar="KUBECONFIG")
    ] = None,
    image: Annotated[str, typer.Option(help="Docker image")] = None,
    registry_pat: Annotated[
        str,
        typer.Option(
            help="Registry personal access token (PAT)",
            envvar="REGISTRY_PAT",
        ),
    ] = None,
):
    if image is None:
        project = Project()
        image = project.cached_images(target.value)

    if kubeconfig.exists() is False:
        typer.echo(f"kubeconfig file {kubeconfig} does not exist")
        raise typer.Exit(code=1)

    kubeconfig = os.environ.get("KUBECONFIG", kubeconfig.as_posix())

    if kubeconfig is None:
        typer.echo("KUBECONFIG not set")
        raise typer.Exit(code=1)

    # config.load_kube_config(kubeconfig)

    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        TimeElapsedColumn(),
        expand=True,
    ) as progress:
        k8s_context = K8sContext(kubeconfig, progress=progress)
        k8s_context.set_target(target)
        k8s_context.set_container_image(image)
        k8s_context.set_registry_pat(registry_pat)

        with open(file, "r") as f:
            code = f.read()
            # output, stream_name = execute_k8s(code, None, image, registry_pat)
            output, stream_name = k8s_context.execute(code)

            print(f"output:\n{output}")
            print(f"output stream: {stream_name}")


@app.command()
def jupyter(
    install: Annotated[
        bool,
        typer.Option(
            help="Install kernel spec for Jupyter",
        ),
    ] = False,
):
    print("install:", install)
    if install:
        install_my_kernel_spec(user=False, prefix=sys.prefix)
        # install_my_kernel_spec(user=user, prefix=prefix)


# if __name__ == "__main__":
app()



import os
from typing import Dict
import pluggy
from kubernetes import client

from q8s.constants import WORKSPACE
from q8s.enums import Target

hookspec = pluggy.HookspecMarker("q8s")
hookimpl = pluggy.HookimplMarker("q8s")


class JobTemplatePluginSpec:

    @hookspec
    def prepare(
        self,
        target: Target,
        name: str,
        namespace: str,
        env: Dict[
            str,
            str | None,
        ],
    ) -> None:
        pass

    @hookspec
    def makejob(
        self,
        name: str,
        registry_pat: str | None,
        registry_credentials_secret_name: str,
        container_image: str,
        env: Dict[
            str,
            str | None,
        ],
        target: Target,
    ) -> client.V1PodTemplateSpec:
        return None

    @hookspec
    def cleanup(self, name: str, namespace: str) -> None:
        pass



from os import remove
from os.path import exists, join
import unittest
from unittest.mock import Mock
from rich.progress import Progress, SpinnerColumn

from q8s.project import Project


class TestProject(unittest.TestCase):
    def before(self):
        pass

    def after(self):
        remove("tests/fixtures/cache/.q8s_cache/cpu/requirements.txt", missing_ok=True)

    def test_init(self):

        project = Project("tests/fixtures/cache")

        self.assertEqual(project.name, "Example")
        self.assertIsNotNone(project.configuration.python_env)
        self.assertEqual(project.configuration.targets.keys(), {"cpu", "gpu"})

    def test_init_cache(self):
        project_path = "tests/fixtures/cache"
        project = Project(project_path)

        self.assertFalse(exists(join(project_path, ".q8s_cache/cpu/requirements.txt")))
        self.assertFalse(exists(join(project_path, ".q8s_cache/gpu/requirements.txt")))

        project.init_cache()

        self.assertTrue(exists(join(project_path, ".q8s_cache/cpu/requirements.txt")))
        self.assertTrue(exists(join(project_path, ".q8s_cache/gpu/requirements.txt")))

    def test_clear_cache(self):
        project_path = "tests/fixtures/cache"
        project = Project(project_path)

        project.init_cache()

        self.assertTrue(exists(join(project_path, ".q8s_cache/cpu/requirements.txt")))
        self.assertTrue(exists(join(project_path, ".q8s_cache/gpu/requirements.txt")))

        project.clear_cache()

        self.assertFalse(exists(join(project_path, ".q8s_cache/cpu/requirements.txt")))
        self.assertFalse(exists(join(project_path, ".q8s_cache/gpu/requirements.txt")))

    def test_check_cache(self):
        project_path = "tests/fixtures/cache"
        project = Project(project_path)

        project.init_cache()

        with open(join(project_path, ".q8s_cache/cpu/requirements.txt"), "a") as f:
            f.write("qiskit==1.0.0")

        self.assertFalse(project.check_cache())

        project.clear_cache()

    @unittest.mock.patch("os.system", return_value=0)
    def test_build_container(self, mock_system: Mock):
        project_path = "tests/fixtures/cache"
        project = Project(project_path)

        project.init_cache()

        image_name = project.build_container(
            target="cpu", progress=Progress(SpinnerColumn()), push=False
        )

        assert mock_system.called
        assert (
            mock_system.call_args[0][0]
            == f"docker build -t {image_name} {join(project_path, '.q8s_cache/cpu')}"
        )
        self.assertEqual(image_name, "vstirbu/q8s-example:cpu")

    @unittest.mock.patch("os.system", return_value=0)
    def test_push_container(self, mock_system: Mock):
        project_path = "tests/fixtures/cache"
        project = Project(project_path)

        project.init_cache()

        image_name = project.build_container(
            target="cpu", progress=Progress(SpinnerColumn()), push=False
        )

        project.push_container(target="cpu", progress=Progress(SpinnerColumn()))

        assert mock_system.called
        assert mock_system.call_args[0][0] == f"docker push {image_name}"



import os
from ipykernel.kernelbase import Kernel
from rich.progress import Progress, SpinnerColumn, TextColumn
import logging

from q8s.execution import K8sContext

FORMAT = "[%(levelname)s %(asctime)-15s q8s_kernel] %(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT)

CODE = ["test_function", "import json; json.dumps(test_function)"]


class Q8sKernel(Kernel):
    implementation = "q8s-kernel"
    implementation_version = "0.1.0"
    language = "no-op"
    language_version = "0.1"
    language_info = {
        "name": "Any text",
        "mimetype": "text/x-python",
        "pygments_lexer": "ipython%d" % 3,
        "file_extension": ".py",
        "nbconvert_exporter": "python",
    }
    banner = "q8s"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.docker_image = os.environ.get("DOCKER_IMAGE", "vstirbu/benchmark-deps")
        kubeconfig = os.environ.get("KUBECONFIG", None)

        if kubeconfig is None:
            logging.error("KUBECONFIG not set")
            exit(1)

        self.k8s_context = K8sContext(
            kubeconfig,
            self.progress,
            Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                expand=True,
            ),
        )
        self.k8s_context.set_container_image(self.docker_image)
        self.k8s_context.set_registry_pat(os.environ.get("REGISTRY_PAT", None))

        logging.info("q8s kernel started")
        logging.info(f"docker image: {self.docker_image}")

    def do_execute(
        self,
        code,
        silent,
        store_history=True,
        user_expressions=None,
        allow_stdin=False,
    ):
        logging.info(f"Executing code:\n{code}")

        output, stream_name = self.k8s_context.execute(code)

        self.send_response(self.iopub_socket, "clear_output", {"wait": True})

        logging.debug(output)
        logging.debug(stream_name)

        for line in output.split("\n"):
            if line.startswith("data:image/png;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/png": line[22:]},
                        "metadata": {},
                    },
                )
            elif line.startswith("data:image/jpeg;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/jpeg": line[23:]},
                        "metadata": {},
                    },
                )
            elif line.startswith("data:image/svg+xml;base64,"):
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"image/svg+xml": line[25:]},
                        "metadata": {},
                    },
                )
            else:
                self.send_response(
                    self.iopub_socket,
                    "display_data",
                    {
                        "data": {"text/plain": line},
                        "metadata": {},
                    },
                )

        return {
            "status": "ok",
            # The base class increments the execution count
            "execution_count": self.execution_count,
            "payload": [],
            "user_expressions": {},
        }

    def progress(self, msg):
        self.send_response(
            self.iopub_socket,
            "display_data",
            {
                "data": {"text/plain": msg},
                "metadata": {},
            },
        )
